from typing import List
import torch
from PIL import Image
from surya.settings import settings
from tqdm import tqdm
import numpy as np


def get_batch_size():
    batch_size = settings.RECOGNITION_BATCH_SIZE
    if batch_size is None:
        batch_size = 32
        if settings.TORCH_DEVICE_MODEL == "mps":
            batch_size = 64 # 12GB RAM max
        if settings.TORCH_DEVICE_MODEL == "cuda":
            batch_size = 256
    return batch_size


def batch_recognition(images: List, languages: List[List[str]], model, processor):
    assert all([isinstance(image, Image.Image) for image in images])
    assert len(images) == len(languages)
    batch_size = get_batch_size()

    images = [image.convert("RGB") for image in images]

    output_text = []
    for i in tqdm(range(0, len(images), batch_size), desc="Recognizing Text"):
        batch_langs = languages[i:i+batch_size]
        batch_images = images[i:i+batch_size]
        model_inputs = processor(text=[""] * len(batch_langs), images=batch_images, lang=batch_langs)

        batch_pixel_values = model_inputs["pixel_values"]
        batch_langs = model_inputs["langs"]
        batch_decoder_input = [[model.config.decoder_start_token_id] + lang for lang in batch_langs]

        batch_langs = torch.from_numpy(np.array(batch_langs, dtype=np.int64)).to(model.device)
        batch_pixel_values = torch.tensor(np.array(batch_pixel_values), dtype=model.dtype).to(model.device)
        batch_decoder_input = torch.from_numpy(np.array(batch_decoder_input, dtype=np.int64)).to(model.device)

        with torch.inference_mode():
            generated_ids = model.generate(
                pixel_values=batch_pixel_values,
                decoder_input_ids=batch_decoder_input,
                decoder_langs=batch_langs,
                eos_token_id=processor.tokenizer.eos_id,
                max_new_tokens=settings.RECOGNITION_MAX_TOKENS
            )

        output_text.extend(processor.tokenizer.batch_decode(generated_ids))

    return output_text









import os

import requests
from PIL import Image, ImageDraw, ImageFont
from surya.settings import settings


def get_text_size(text, font):
    im = Image.new(mode="P", size=(0, 0))
    draw = ImageDraw.Draw(im)
    _, _, width, height = draw.textbbox((0, 0), text=text, font=font)
    return width, height


def draw_text_on_image(bboxes, texts, image_size=(1024, 1024), font_path=settings.RECOGNITION_RENDER_FONT, max_font_size=60, res_upscale=2):
    new_image_size = (image_size[0] * res_upscale, image_size[1] * res_upscale)
    image = Image.new('RGB', new_image_size, color='white')
    draw = ImageDraw.Draw(image)

    for bbox, text in zip(bboxes, texts):
        s_bbox = [coord * res_upscale for coord in bbox]
        bbox_width = s_bbox[2] - s_bbox[0]
        bbox_height = s_bbox[3] - s_bbox[1]

        # Shrink the text to fit in the bbox if needed
        box_font_size = min(int(.75 * bbox_height), max_font_size)

        # Download font if it doesn't exist
        if not os.path.exists(font_path):
            with requests.get(settings.RECOGNITION_FONT_DL_PATH, stream=True) as r, open(font_path, 'wb') as f:
                r.raise_for_status()
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        font = ImageFont.truetype(font_path, box_font_size)
        text_width, text_height = get_text_size(text, font)
        while (text_width > bbox_width or text_height > bbox_height) and box_font_size > 6:
            box_font_size = box_font_size - 1
            font = ImageFont.truetype(font_path, box_font_size)
            text_width, text_height = get_text_size(text, font)

        # Calculate text position (centered in bbox)
        text_width, text_height = get_text_size(text, font)
        x = s_bbox[0]
        y = s_bbox[1] + (bbox_height - text_height) / 2

        draw.text((x, y), text, fill="black", font=font)

    return image



from collections import defaultdict
from typing import List
from tqdm import tqdm

import torch
from PIL import Image

from surya.detection import batch_detection
from surya.input.processing import slice_polys_from_image, slice_bboxes_from_image
from surya.recognition import batch_recognition


def run_recognition(images: List[Image.Image], langs: List[List[str]], rec_model, rec_processor, bboxes: List[List[List[int]]] = None, polygons: List[List[List[List[int]]]] = None):
    # Polygons need to be in corner format - [[x1, y1], [x2, y2], [x3, y3], [x4, y4]], bboxes in [x1, y1, x2, y2] format
    assert bboxes is not None or polygons is not None
    slice_map = []
    all_slices = []
    all_langs = []
    for idx, (image, lang) in enumerate(zip(images, langs)):
        if polygons is not None:
            slices = slice_polys_from_image(image, polygons[idx])
        else:
            slices = slice_bboxes_from_image(image, bboxes[idx])
        slice_map.append(len(slices))
        all_slices.extend(slices)
        all_langs.extend([lang] * len(slices))

    rec_predictions = batch_recognition(all_slices, all_langs, rec_model, rec_processor)

    predictions_by_image = []
    slice_start = 0
    for idx, (image, lang) in enumerate(zip(images, langs)):
        slice_end = slice_start + slice_map[idx]
        image_lines = rec_predictions[slice_start:slice_end]
        slice_start = slice_end

        pred = {
            "text_lines": image_lines,
            "language": lang
        }

        if polygons is not None:
            pred["polys"] = polygons[idx]
        else:
            pred["bboxes"] = bboxes[idx]

        predictions_by_image.append(pred)

    return predictions_by_image


def run_ocr(images: List[Image.Image], langs: List[List[str]], det_model, det_processor, rec_model, rec_processor):
    det_predictions = batch_detection(images, det_model, det_processor)
    if det_model.device == "cuda":
        torch.cuda.empty_cache() # Empty cache from first model run

    slice_map = []
    all_slices = []
    all_langs = []
    for idx, (image, det_pred, lang) in enumerate(zip(images, det_predictions, langs)):
        slices = slice_polys_from_image(image, det_pred["polygons"])
        slice_map.append(len(slices))
        all_slices.extend(slices)
        all_langs.extend([lang] * len(slices))

    rec_predictions = batch_recognition(all_slices, all_langs, rec_model, rec_processor)

    predictions_by_image = []
    slice_start = 0
    for idx, (image, det_pred, lang) in enumerate(zip(images, det_predictions, langs)):
        slice_end = slice_start + slice_map[idx]
        image_lines = rec_predictions[slice_start:slice_end]
        slice_start = slice_end

        assert len(image_lines) == len(det_pred["polygons"]) == len(det_pred["bboxes"])
        predictions_by_image.append({
            "text_lines": image_lines,
            "polys": det_pred["polygons"],
            "bboxes": det_pred["bboxes"],
            "language": lang
        })

    return predictions_by_image



import subprocess
import os


def run_app():
    cur_dir = os.path.dirname(os.path.abspath(__file__))
    ocr_app_path = os.path.join(cur_dir, "ocr_app.py")
    subprocess.run(["streamlit", "run", ocr_app_path])


import copy
from typing import List, Tuple

from pydantic import BaseModel, field_validator, computed_field


class PolygonBox(BaseModel):
    corners: List[List[float]]

    @field_validator('corners')
    @classmethod
    def check_elements(cls, v: List[List[float]]) -> List[List[float]]:
        if len(v) != 4:
            raise ValueError('corner must have 4 elements')

        for corner in v:
            if len(corner) != 2:
                raise ValueError('corner must have 2 elements')
        return v

    @property
    def height(self):
        return self.corners[1][1] - self.corners[0][1]

    @property
    def width(self):
        return self.corners[1][0] - self.corners[0][0]

    @property
    def area(self):
        return self.width * self.height

    @computed_field
    @property
    def bbox(self) -> List[float]:
        box = [self.corners[0][0], self.corners[0][1], self.corners[1][0], self.corners[2][1]]
        if box[0] > box[2]:
            box[0], box[2] = box[2], box[0]
        if box[1] > box[3]:
            box[1], box[3] = box[3], box[1]
        return box


    def rescale(self, processor_size, image_size):
        # Point is in x, y format
        page_width, page_height = processor_size

        img_width, img_height = image_size
        width_scaler = img_width / page_width
        height_scaler = img_height / page_height

        new_corners = copy.deepcopy(self.corners)
        for corner in new_corners:
            corner[0] = int(corner[0] * width_scaler)
            corner[1] = int(corner[1] * height_scaler)
        self.corners = new_corners



class Bbox(BaseModel):
    bbox: List[float]

    @field_validator('bbox')
    @classmethod
    def check_4_elements(cls, v: List[float]) -> List[float]:
        if len(v) != 4:
            raise ValueError('bbox must have 4 elements')
        return v

    @property
    def height(self):
        return self.bbox[3] - self.bbox[1]

    @property
    def width(self):
        return self.bbox[2] - self.bbox[0]

    @property
    def area(self):
        return self.width * self.height


from typing import List

import cv2
import torch
import numpy as np
from PIL import Image
from surya.postprocessing.heatmap import get_and_clean_boxes
from surya.postprocessing.affinity import get_vertical_lines, get_horizontal_lines
from surya.input.processing import prepare_image, split_image
from surya.settings import settings
from tqdm import tqdm


def get_batch_size():
    batch_size = settings.DETECTOR_BATCH_SIZE
    if batch_size is None:
        batch_size = 8
        if settings.TORCH_DEVICE_MODEL == "cuda":
            batch_size = 32
    return batch_size


def batch_detection(images: List, model, processor):
    assert all([isinstance(image, Image.Image) for image in images])
    batch_size = get_batch_size()

    images = [image.convert("RGB") for image in images]
    orig_sizes = [image.size for image in images]
    split_index = []
    split_heights = []
    image_splits = []
    for i, image in enumerate(images):
        image_parts, split_height = split_image(image, processor)
        image_splits.extend(image_parts)
        split_index.extend([i] * len(image_parts))
        split_heights.extend(split_height)

    image_splits = [prepare_image(image, processor) for image in image_splits]

    pred_parts = []
    for i in tqdm(range(0, len(image_splits), batch_size), desc="Detecting bboxes"):
        batch = image_splits[i:i+batch_size]
        # Batch images in dim 0
        batch = torch.stack(batch, dim=0)
        batch = batch.to(model.dtype)
        batch = batch.to(model.device)

        with torch.inference_mode():
            pred = model(pixel_values=batch)

        logits = pred.logits
        for j in range(logits.shape[0]):
            heatmap = logits[j, 0, :, :].detach().cpu().numpy().astype(np.float32)
            affinity_map = logits[j, 1, :, :].detach().cpu().numpy().astype(np.float32)

            heatmap_shape = list(heatmap.shape)
            correct_shape = [processor.size["height"], processor.size["width"]]
            cv2_size = list(reversed(correct_shape)) # opencv uses (width, height) instead of (height, width)

            if heatmap_shape != correct_shape:
                heatmap = cv2.resize(heatmap, cv2_size, interpolation=cv2.INTER_LINEAR)

            affinity_shape = list(affinity_map.shape)
            if affinity_shape != correct_shape:
                affinity_map = cv2.resize(affinity_map, cv2_size, interpolation=cv2.INTER_LINEAR)

            pred_parts.append((heatmap, affinity_map))

    preds = []
    for i, (idx, height) in enumerate(zip(split_index, split_heights)):
        if len(preds) <= idx:
            preds.append(pred_parts[i])
        else:
            heatmap, affinity_map = preds[idx]
            pred_heatmap = pred_parts[i][0]
            pred_affinity = pred_parts[i][1]

            if height < processor.size["height"]:
                # Cut off padding to get original height
                pred_heatmap = pred_heatmap[:height, :]
                pred_affinity = pred_affinity[:height, :]

            heatmap = np.vstack([heatmap, pred_heatmap])
            affinity_map = np.vstack([affinity_map, pred_affinity])
            preds[idx] = (heatmap, affinity_map)

    assert len(preds) == len(images)
    results = []
    for i in range(len(images)):
        heatmap, affinity_map = preds[i]
        heat_img = Image.fromarray((heatmap * 255).astype(np.uint8))
        aff_img = Image.fromarray((affinity_map * 255).astype(np.uint8))

        affinity_size = list(reversed(affinity_map.shape))
        heatmap_size = list(reversed(heatmap.shape))
        bboxes = get_and_clean_boxes(heatmap, heatmap_size, orig_sizes[i])
        bbox_data = [bbox.model_dump() for bbox in bboxes]
        vertical_lines = get_vertical_lines(affinity_map, affinity_size, orig_sizes[i])
        horizontal_lines = get_horizontal_lines(affinity_map, affinity_size, orig_sizes[i])

        results.append({
            "bboxes": [bbd["bbox"] for bbd in bbox_data],
            "polygons": [bbd["corners"] for bbd in bbox_data],
            "vertical_lines": vertical_lines,
            "horizontal_lines": horizontal_lines,
            "heatmap": heat_img,
            "affinity_map": aff_img,
        })

    return results






